/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Comparator;
import java.util.NoSuchElementException;
import java.util.Objects;

import static java.lang.System.arraycopy;

/**
 * Mutable array of objects.
 *
 * @author Leon van Zantvoort
 */
class MutableObjectArrayImpl<E> extends ObjectArrayImpl<E> implements MutableObjectArray<E>, MutableArrayCharacteristics {

    /** Controls state of array. Is {@code null} if control is transferred to caller of {@code doAction}. */
    transient ArrayContext context;

    /**
     * Package visible constructor.
     *
     * @param array backing array.
     * @param offset start index of range.
     * @param length length of range.
     * @param characteristics array characteristics.
     * @param copy {@code true} to make a defensive copy of {@code array}, {@code false} to use {@code array} reference
     *             as backing array.
     */
    MutableObjectArrayImpl(@Nonnull Object[] array, int offset, int length, int characteristics,
                           @Nullable java.util.Comparator<? super E> comparator, boolean copy) {
        super(array, offset, length, characteristics, comparator, copy);
        context = new ArrayContextImpl();
    }

    /**
     * Returns array context, or {@code null} if {@code doAction} is being executed.
     *
     * @return array context, or {@code null} if {@code doAction} is being executed.
     */
    protected ArrayContext getArrayContext() {
        return context;
    }

    /**
     * Updates the SORTED and DISTINCT characteristics of this array by looking at the elements surrounding the
     * specified {@code value}. The SORTED and/or DISTINCT characteristics are removed if the value invalidates the
     * characteristics.
     *
     * @param elementIndex index relative to the backing array.
     * @param object object that is stored at index.
     */
    @SuppressWarnings("unchecked")
    protected void updateCharacteristics(int elementIndex, E object) {
        assert hasCharacteristics(SORTED);
        if (length > 1) {
            Comparator<Object> c = null;
            if (comparator == null) {
                if (object != null && object instanceof Comparable) {
                    c = Arrays.NATURAL_OBJECT_ORDERING;
                } else {
                    removeCharacteristics(SORTED | DISTINCT);
                }
            } else {
                c = (Comparator<Object>) comparator;
            }
            if (c != null) {
                if (elementIndex > offset) {
                    if (c.compare(object, elements[elementIndex - 1]) < 0) {
                        removeCharacteristics(SORTED | DISTINCT);
                    } else if (c.compare(object, elements[elementIndex - 1]) == 0) {
                        removeCharacteristics(DISTINCT);
                    }
                }
                if (elementIndex < offset + length - 1) {
                    if (c.compare(object, elements[elementIndex + 1]) > 0) {
                        removeCharacteristics(SORTED | DISTINCT);
                    } else if (c.compare(object, elements[elementIndex + 1]) == 0) {
                        removeCharacteristics(DISTINCT);
                    }
                }
            }
        }
    }

    /**
     * Updates the SORTED and DISTINCT characteristics of this array by looking at the elements surrounding the
     * first and last element of the specified {@code array}. The SORTED and/or DISTINCT characteristics are removed if
     * these elements invalidate the characteristics.
     *
     * @param elementIndex index relative to the backing array.
     * @param array array that is stored, starting at index.
     */
    @SuppressWarnings("unchecked")
    protected void updateCharacteristics(int elementIndex, @Nonnull Array<? extends E> array) {
        assert hasCharacteristics(SORTED);
        int len = array.length();
        if (length > len) {
            if (len > 0) {
                updateCharacteristics(elementIndex, (E) elements[elementIndex]);
            }
            if (len > 1) {
                updateCharacteristics(elementIndex + len - 1, (E) elements[elementIndex + len - 1]);
            }
        }
    }

    /**
     * Sets the specified characteristics for this array.
     *
     * @param characteristics characteristics to be set for this array.
     * @return the characteristics that were actually stored for this array.
     */
    @Override
    public int characteristics(int characteristics) {
        if (characteristics != this.characteristics) { // Optimization.
            ArrayContext context = getArrayContext();
            if (context != null) {
                return context.characteristics(characteristics);
            }
        }
        return this.characteristics;
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    @Override
    protected int filterCharacteristics(int characteristics) {
        return characteristics | MUTABLE & ~DYNAMIC;
    }

    /**
     * Performs the specified {@code action} without affecting the array's set of characteristics. Ownership of
     * array characteristics control is moved to the caller by means of the {@code ArrayContext} object. Errors or
     * runtime exceptions thrown by the action are relayed to the caller.
     *
     * @param <T> return type.
     * @param action the action to be performed.
     * @return the value returned by the action's {@code run} method.
     * @throws IllegalStateException if action is called when array does not own control of characteristics.
     * @exception NullPointerException if the action is {@code null}.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public final <T> T doAction(@Nonnull ArrayAction<T> action) {
        ArrayContext context = getArrayContext();
        if (context == null) {
            throw new IllegalStateException();
        }
        try {
            this.context = null;
            return action.run(context);
        } finally {
            this.context = context;
        }
    }

    /**
     * Sets the comparator that is used for sorting this array.
     *
     * @param comparator that is used for sorting this array.
     */
    protected void setComparator(@Nullable Comparator<? super E> comparator) {
        this.comparator = comparator;
    }

    /**
     * Replaces element at specified {@code index} by specified object.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code object} is not {@code null};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index index of item to be replaced, relative to the offset of the array.
     * @param object object to replace existing object.
     * @return object that is replaced.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @SuppressWarnings({"unchecked", "InconsistentJavaDoc"})
    @Override
    public @CheckForNull E set(int index, @Nullable E object) {
        rangeCheck(index);
        int elementIndex = offset + index;
        E oldObject = (E) elements[elementIndex];
        elements[elementIndex] = object;
        if (object == null) {
            removeCharacteristics(NONNULL);
        }
        if (hasCharacteristics(SORTED)) {
            updateCharacteristics(elementIndex, object);
        }
        return oldObject;
    }

    /**
     * Replaces all elements of this array by the elements of the specified array. Note that the length of {@code array}
     * needs to be equal to the length of {@code this} array, throws an {@code IllegalArgumentException} otherwise.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED};</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param array array to replace elements of this array.
     * @throws IllegalArgumentException if length of specified array is not equal to length of this array.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void setAll(@Nonnull Array<? extends E> array) {
        int lengthDiff = array.length() - length;
        if (lengthDiff == 0) {
            arraycopy(UnsafeArray.readAccess(array), array.offset(), elements, offset, length);
            characteristics(array.characteristics());   // All elements are replaced by new array; Copy characteristics.
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(offset, array);
            }
        } else {
            throw new IllegalArgumentException(String.format("Array length (%d) not equal to length of this array (%d).",
                    array.length(), length));
        }
    }

    /**
     * Sorts this array according to natural ordering.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     */
    @Override
    public void sort() {
        if (!hasCharacteristics(SORTED) || !Objects.equals(comparator, null)) {
            java.util.Arrays.sort(elements, offset, offset + length, Arrays.NATURAL_OBJECT_ORDERING);
            setComparator(null);
            addCharacteristics(SORTED);
        }
    }

    /**
     * Sorts this array according to natural ordering. The underlying implementation utilizes multiple threads
     * to reduce sorting time. The algorithm may require a working space equal to the size of this array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     */
    @Override
    public void parallelSort() {
        if (!hasCharacteristics(SORTED) || !Objects.equals(comparator, null)) {
            java.util.Arrays.parallelSort(elements, offset, offset + length, Arrays.NATURAL_OBJECT_ORDERING);
            setComparator(null);
            addCharacteristics(SORTED);
        }
    }

    /**
     * Sorts this array using the supplied {@code comparator} to compare elements. The {@code comparator} can be
     * retrieved by {@code getComparator} after this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     *
     * @param comparator the {@code Comparator} used to compare list elements.
     *          A {@code null} value indicates that the elements'
     *          {@linkplain Comparable natural ordering} should be used
     * @throws ClassCastException if the list contains elements that are not
     *         <i>mutually comparable</i> using the specified comparator
     * @throws IllegalArgumentException
     *         (<a href="Collection.html#optional-restrictions">optional</a>)
     *         if the comparator is found to violate the {@link java.util.Comparator}
     *         contract
     * @throws NullPointerException if the specified comparator is null.
     */
    @SuppressWarnings("unchecked")
    @Override
    public void sort(Comparator<? super E> comparator) {
        if (!hasCharacteristics(SORTED) || !Objects.equals(this.comparator, comparator)) {
            java.util.Arrays.sort(elements, offset, offset + length, (Comparator<Object>) comparator);
            setComparator(comparator);
            addCharacteristics(SORTED);
        }
    }

    /**
     * Sorts this array using the supplied {@code Comparator} to compare elements. The underlying implementation
     * utilizes multiple threads to reduce sorting time. The algorithm requires a working space equal to the size of the
     * this array. The {@code comparator} can be retrieved by {@code getComparator} after this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     *
     * @param comparator the {@code Comparator} used to compare array elements.
     *          A {@code null} value indicates that the elements'
     *          {@linkplain Comparable natural ordering} should be used
     * @throws ClassCastException if the list contains elements that are not
     *         <i>mutually comparable</i> using the specified comparator
     * @throws IllegalArgumentException
     *         (<a href="Collection.html#optional-restrictions">optional</a>)
     *         if the comparator is found to violate the {@link java.util.Comparator}
     *         contract
     * @throws NullPointerException if the specified comparator is null.
     */
    @SuppressWarnings("unchecked")
    @Override
    public void parallelSort(Comparator<? super E> comparator) {
        if (!hasCharacteristics(SORTED) || !Objects.equals(this.comparator, comparator)) {
            java.util.Arrays.parallelSort(elements, offset, offset + length, (Comparator<Object>) comparator);
            setComparator(comparator);
            addCharacteristics(SORTED);
        }
    }

    /**
     * Array context implementation.
     */
    @SuppressWarnings("MissingMethodJavaDoc")
    private class ArrayContextImpl implements ArrayContext {
        @Override
        public int characteristics(int newCharacteristics) {
            characteristics = filterCharacteristics(newCharacteristics);
            return characteristics;
        }

        @Override
        public int characteristics() {
            return characteristics;
        }
    }

    /**
     * Sub arrays share serialization proxy with parent arrays intentionally. As a result, deserialized sub arrays become
     * detached from their originating parent.
     */
    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SubMutableObjectArrayImpl<E> extends MutableObjectArrayImpl<E> {

        final MutableObjectArrayImpl<E> parent;

        @SuppressWarnings("unchecked")
        SubMutableObjectArrayImpl(@Nonnull MutableObjectArrayImpl<E> parent, int offset, int length) {
            super(parent.elements, offset, length, parent.characteristics, parent.comparator, false);
            this.parent = parent;
        }

        @Override
        protected ArrayContext getArrayContext() {
            return parent.getArrayContext() == null ? null : context;
        }

        @Override
        protected void updateCharacteristics(int elementIndex, E object) {
            super.updateCharacteristics(elementIndex, object);
            if (hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, object);
            }
        }

        @Override
        protected void updateCharacteristics(int elementIndex, @Nonnull Array<? extends E> array) {
            super.updateCharacteristics(elementIndex, array);
            if (hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, array);
            }
        }

        @Override
        public int characteristics(int characteristics) {
            // Parent array retains characteristics that are shared with sub array. Additional characteristics
            // assigned to sub array are not set for parent array, as it is not guaranteed that these
            // characteristics hold for parent array.
            parent.retainCharacteristics(characteristics);
            return super.characteristics(characteristics);
        }

        @Override
        public void setComparator(@Nullable Comparator<? super E> comparator) {
            super.setComparator(comparator);
            parent.setComparator(comparator);
        }
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array. The sub array has the same mutability properties as {@code this} array.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public MutableObjectArray<E> offset(int offset) {
        if (offset == 0) {
            return this;
        }
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new SubMutableObjectArrayImpl<>(this, newOffset, newLength);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    public MutableObjectArray<E> length(int length) {
        if (this.length == length) {
            return this;
        }
        subArrayCheck(offset, length);
        return new SubMutableObjectArrayImpl<>(this, offset, length);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    public MutableObjectArray<E> subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        if (newOffset == offset && newLength == length) {
            return this;
        }
        subArrayCheck(newOffset, newLength);
        return new SubMutableObjectArrayImpl<>(this, newOffset, newLength);
    }


    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator<E> {
        int index;
        int lastIndex = -1;

        ArrayIteratorImpl() {
            this(0);
        }

        ArrayIteratorImpl(int index) {
            this.index = offset + index;
        }

        @Override
        public boolean hasNext() {
            return index < offset + length;
        }

        @SuppressWarnings("unchecked")
        @Override
        public E next() {
            int i = index;
            if (i >= offset + length) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i + 1;
            return (E) elements[lastIndex = i];
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return index > offset;
        }

        @SuppressWarnings("unchecked")
        @Override
        public E previous() {
            int i = index - 1;
            if (i < offset) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i;
            return (E) elements[lastIndex = i];
        }

        @Override
        public int nextIndex() {
            return index;
        }

        @Override
        public int previousIndex() {
            return index - 1;
        }

        @Override
        public void set(E object) {
            if (lastIndex < 0) {
                throw new IllegalStateException();
            }
            assert lastIndex < offset + length && lastIndex < elements.length;
            MutableObjectArrayImpl.this.set(lastIndex, object);
        }

        @Override
        public void add(E value) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator<E> iterator() {
        return new ArrayIteratorImpl();
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator<E> iterator(int index) {
        if (index > length) {
            throw new IndexOutOfBoundsException();
        }
        return new ArrayIteratorImpl(index);
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    @Override
    Object writeReplace() {
        return new SerializationProxy<>(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy<E> implements Serializable {
        static final long serialVersionUID = 291623392563940252L;

        transient int offset;
        final int length;
        final int characteristics;
        final Comparator<? super E> comparator;
        final Class<?> cls;
        transient Object[] elements;

        SerializationProxy(MutableObjectArrayImpl<E> array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.comparator = array.comparator;
            this.elements = array.array();
            this.cls = this.elements.getClass().getComponentType();
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            for (int i = 0; i < length; i++) {
                s.writeObject(elements[offset + i]);
            }
        }

        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = (Object[]) java.lang.reflect.Array.newInstance(cls, length);
            for (int i = 0; i < length; i++) {
                elements[i] = s.readObject();
            }
        }

        @SuppressWarnings("unchecked")
        private Object readResolve() {
            return new MutableObjectArrayImpl<>(elements, 0, length, characteristics, comparator, false);
        }
    }
}
